package SAPDB::Install::Templates2::InstRegistry;
use SAPDB::Install::Templates2::Properties;
use SAPDB::Install::Templates2::Util;
use SAPDB::Install::Templates2::Debug;
use SAPDB::Install::Registry;
use SAPDB::Install::Templates2::InstUnit;
use strict;

our @ISA;
if($SCRIPTING){
  require Exporter;
  push @ISA,qw(Exporter);
}
else{
  require SDB::Common::Exporter;
  push @ISA,qw(SDB::Common::Exporter);
}
push @ISA , qw(SAPDB::Install::Templates2::Properties);



# definition of package registry structure
my %RegStruct = (
			ID => 'SAP DB Package Registry',		
			Version => '0.1',
			HashRef_Packages => {},				
			InstanceData => undef					
);


sub readInstReg{
  my ($aFile)=@_; 
  my $dump_string;
  my $retVal=0;
  my $pos;
  open(dumpFD,$aFile) or return("Cannot open registry file   \"$aFile\"\n");
  while (<dumpFD>) {
    $dump_string.=$_;
  }	
  close(dumpFD);

  my $counts=0;
  $dump_string =~ s/(0*$)//s and $counts = length($1);
  my $struct_ref=getElement($dump_string);
  return $struct_ref;
}


sub new{
  my $file = shift;
  my $keys={
            'data' => {},
            'lastError' => {},
            'location' => {}
           };
  my $this = {'KEYS' => $keys};
  bless $this;
  my $data;
  if ($file) {
    $data = readInstReg($file);
    $this->setLocation($file);
    if (ref $data) {
      $this->setData($data);
    } else {
      print STDERR $data;
      return undef;
    }
  } else {
    my %newData = %RegStruct;
    $data = \%newData;
    $this->setData($data);
  }
  return  $this;
}


# normalize windows path strings: lower case and slashes  
# return 0 if nothing changed else 1
sub normalize{
	my ($string)=@_;
	$^O=~/mswin/i or return $string;
	my $path=$string;
	$path=~s/\\/\//g; 
	$path=~tr/A-Z/a-z/;
	return $path;	
}



sub getElement{
  # sub parse dumpstring and create struct elements in memory
  my ($dump_string)=@_;
  my @returnvalue;
  $_=$dump_string;
  if (/^ARRAY\((.*)\)$/s) {
    my @emptyArray;
    $1 eq '' and return \@emptyArray;
    my @array=splitParseString($1);
    foreach my $element (@array) {
      my $anything=getElement($element);
      push @returnvalue,$anything;
    }
    return \@returnvalue;
  } elsif (/^HASH\((.*)\)$/s) {
    my %emptyHash;
    $1 eq '' and return \%emptyHash;
    my @array=splitParseString($1);
    my %returnhash;
    foreach my $element (@array) {
      $_=$element;
      /^""(.*?)""\s*=>\s*(\S.*)$/s or print("Array Read error in registry: '$element'\n") and return undef;
      my $key=demask($1);
      my $anything=getElement($2);
      $returnhash{$key}=$anything;
    }
    return \%returnhash;
  } elsif (/^REF\((.*)\)$/s) {
    my $anything=getElement($1);
    return \$anything;
  } elsif (/^SCALAR\(""(.*)""\)$/s) {
    my $scalar_value=demask($1);
    return \$scalar_value;
  } elsif (/^""(.*)""$/s) {
    my $value=demask($1);
    return $value;
  } elsif (/^undef$/g) {
    return undef;
  } else {
    print("Read error in registry: Unknown type '$_'.\n");
    return undef;
  }
}


sub splitParseString{
  my ($parse_string)=@_;
  my @returnlist;
  my @tmp_array=split(';;',$1);
  my $cons;
  my $cnt1=0;                   # number of "("
  my $cnt2=0;                   # number of ")"
  foreach my $element (@tmp_array) {
    $cnt1+=$element=~s/\(/\(/g; # count "("
    $cnt2+=$element=~s/\)/\)/g; # count ")"
    $cnt1-=$element=~s/(\\\()/$1/g; # minus masked "("
    $cnt2-=$element=~s/(\\\))/$1/g; # minus masked ")"
    my $sep;
    if (not defined $cons or  $cons eq '') {
      $sep='';
    } else {
      $sep=';;';
    }
    $cons.="$sep$element";
    if ($cnt1==$cnt2) {
      push @returnlist,$cons;
      $cons='';
      $cnt1=0;
      $cnt2=0;
    }
  }
  if ($cons) {
    print("Error in splitParseString");
    return undef;
  }
  return @returnlist;
}


sub mask{
  my $real_string=shift;
  $real_string=~s/\(/\\\(/g ; # ( -> \(     delimiter for struct types ARRAY/HASH/REF/SCALAR
  $real_string=~s/\)/\\\)/g ; # ) -> \)     delimiter for struct types ARRAY/HASH/REF/SCALAR
  $real_string=~s/"/\\"/g ; # " -> \"     delimiter for values (scalars) and hash keys ""         
  $real_string=~s/;/\\;/g ; # ; -> \;     seperator for array and hash elements ";;" 
  return $real_string;
}



sub demask{
  my $real_string=shift;
  $real_string=~s/\\;/;/g ; # \; -> ;
  $real_string=~s/\\"/"/g ; # \" -> "
  $real_string=~s/\\\)/\)/g ; # \) -> )
  $real_string=~s/\\\(/\(/g ; # \( -> (
  #$real_string=~s/\\\\/\\/g ;  # \\ -> \
  return $real_string;
}


sub GetAllPackageObjects{
  my ($this,$instConfig) = @_;
  my @allPacks;
  my @packages = $this->GetPackages();
  foreach(@packages){
    my $name = @$_[0];
    my $path = @$_[1];
    my $packobj = $this->createPackageObject($name,$path,$instConfig);
    push @allPacks,$packobj;
  }
  return @allPacks;
}

sub GetPackages{
  my ($this) = @_;
  my $data = $this->getData();
  my $hrRegistryData=$data->{'HashRef_Packages'};
  my @retVal;
  foreach my $packname (keys(%$hrRegistryData)) {
    foreach my $path (keys(%{${%$hrRegistryData}{$packname}})) {
      push @retVal,([$packname,$path,$this->GetVersion($packname,$path)]); 
    }
  }
  return @retVal;
}

sub GetVersion{
  my ($this,$packname,$path) = @_;
  return $this->GetPackageValue($packname,$path,'Version');
}


sub GetFileList{
  my ($this,$packname,$path) = @_;
  return $this->GetPackageValue($packname,$path,'FileList');
}


sub GetPackageValue{
  my ($this,$packname,$path,$key) = @_;
  my $data = $this->getData();
  $path = normalize($path);
  my $hrRegistryData=$data->{'HashRef_Packages'};
#  my $value = ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{$key};
  my ($apack,$apackpath,$value);
  unless (exists $hrRegistryData->{$packname}){ return undef;}
  $apack = $hrRegistryData->{$packname};
  unless (exists $apack->{$path}){ return undef;}
  $apackpath = $apack->{$path};
  unless (exists $apackpath->{$key}){ return undef;}
  $value = $apackpath->{$key};
  return $value;
}

sub SetPackageValue{
  my ($this,$packname,$path,$key,$value) = @_;
  my $data = $this->getData();
  $path = normalize($path);
  my $hrRegistryData=$data->{'HashRef_Packages'};
  my $pkg = $hrRegistryData->{$packname};
  $pkg or return undef;
  my  $pkgp = $pkg->{$path};
  $pkgp or return undef;
  $pkgp->{$key}=$value;
  return $value;
}



sub showInstallations{
  my ($this,$out) = @_;
  unless (defined $out){
    $out = *STDOUT;
  }
  my $data = $this->getData();
  my $text='';
  my $hrRegistryData=$data->{'HashRef_Packages'};
  foreach my $packname (keys(%$hrRegistryData)) {
    print $out "Packagename  '$packname'\n\n";
    foreach my $path (keys(%{${%$hrRegistryData}{$packname}})) {
      my $user = ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'User'};
      my $group = ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'Group'};
      my $packversion = ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'PackageVersion'};
      my $mode = ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'Mode'}; 
      my $pathRef = ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'Pathes'};
      my @paths;
      if ($pathRef) {
        @paths = @$pathRef;
      }
      $user = $user? $user : '<NOT SET>';
      $group = $group? $group : '<NOT SET>';
      $packversion = $packversion? $packversion : '<NOT SET>';
      $mode = $mode? $mode : '<NOT SET>';
      print $out " MAIN PATH: \t\t\t".$path."\n";
      print $out " SOFTWARE VERSION: \t\t".${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'Version'}."\n";
      print $out " PACKAGE VERSION: \t\t$packversion\n";
      print $out " VALID: \t\t\t\t$text\n";
      print $out " BIT: \t\t\t\t$mode bit\n";
      print $out " FILE OWNER: \t\t\t$user\n";
      print $out " FILE GROUP: \t\t\t$group\n";
      print $out " INSTALLATION DATE:\t\t".${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'Date'}."\n";
      my $reqString = join ',',@{${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'Require'}};
      print $out " REQUIRE:\t\t\t".$reqString."\n";
      $text='YES';
      $text='NO' unless(exists ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'FileList'});
      print $out " FILELIST:\t\t\t$text\n";
      $text='YES';
      $text='NO' unless(exists ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'Script'});
      print $out " SCRIPT:\t\t\t\t$text\n";
      #print "PATHES:\n";
      foreach my $path (@paths) {
        my $pathname=${%$path}{'name'};
        $pathname=~tr/[a-z]/[A-Z]/;
        print $out " $pathname:\t\t".${%$path}{'value'}."\n";
      }
      print $out "\n";
    }
    print $out "\n";
  }
}



sub dumpElement{
  # recursive sub dump struct elements
  my ($struct_ref,$fd)=@_;
  my $dumplen=0;	 	
  if (ref($struct_ref) eq 'REF') {
    print $fd 'REF(';
    $dumplen=dumpElement($$struct_ref,$fd);
    print $fd ')';
    $dumplen += 5;
  } elsif (ref($struct_ref) eq 'HASH') {
    print $fd 'HASH(';
    $dumplen += 5;
    my $first=1;
    foreach my $key (keys(%$struct_ref)) {
      if ($first) {
        $first=0;
      } else {
        print $fd ';;';
        $dumplen += 2;
      }
      my $writekey= mask($key);
      print $fd "\"\"$writekey\"\" => ";
      $dumplen += length($writekey) + 8; 
      $dumplen += dumpElement($struct_ref->{$key},$fd);
    }
    print $fd ')';
    $dumplen++;	
  } elsif (ref($struct_ref) eq 'ARRAY') {
    print $fd 'ARRAY(';
    my $first=1;
    foreach my $element (@$struct_ref) {
      if ($first) {
        $first=0;
      } else {
        print $fd ';;';
        $dumplen += 2;
      }
      $dumplen += dumpElement($element,$fd);
    }
    print $fd ')';
    $dumplen++;
  } elsif (ref($struct_ref) eq 'SCALAR') {
    print $fd 'SCALAR(';
    $dumplen += dumpElement($$struct_ref,$fd);
    print $fd ')';
    $dumplen += 8; 
  } elsif (!ref($struct_ref)) {
    if (defined $struct_ref) {
      $struct_ref = mask($struct_ref);
      print $fd '""'.$struct_ref.'""';
      $dumplen += length($struct_ref) + 4;
    } else {
      print $fd "undef";
      $dumplen += 5;
    }	
  } else {
    print $fd ref($struct_ref);
    $dumplen += length(ref($struct_ref)); 	
  }
  return $dumplen;	
}#dumpElement
# save the whole thing back

sub save{
  my $this = shift;
  my $retVal=0;
  my $data = $this->getData();
  my $location = $this->getLocation();
  open OUT, ">$location" or return $!;
  dumpElement($data,*OUT);
  close OUT;
  return $retVal;
}



sub test{
  my $data = readInstReg(@_);
#  showInstallations($data);
#  dumpElement($data,*STDOUT);
  my $testFile = "c:\\temp\\testreg";
  my $AFILE;
  open $AFILE, ">$testFile" or die "Cannot open file";
  dumpElement($data,$AFILE);
  close $AFILE;
  $data = readInstReg($testFile);
  showInstallations($data);
}




sub newPackage{
	my ($this,$packagename,$path)=@_;
#        print "Adding package $packagename $path\n";
	my $dataroot=$this->getData();
        unless(defined  $dataroot->{'HashRef_Packages'}){
          $dataroot->{'HashRef_Packages'} = {};
        }
	my $packageroot=$dataroot->{'HashRef_Packages'};
	my %newHash;
	my %initHash;
	$path = normalize($path);
	my %emptyHash=('Version' => undef); 
	my $hashref_paths=$packageroot->{$packagename};
	if(ref($hashref_paths) eq 'HASH'){
		$hashref_paths->{$path}=\%emptyHash;	
	}
	else{
		$initHash{$path}=\%emptyHash;
		$packageroot->{$packagename}=\%initHash;
	}
        return \%emptyHash;
}


sub generateDate{
  my ($sec,$min,$our,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
  $year = 1900 + $year;
  my $dateString = sprintf "\%02u.\%02u.\%04u" , ($mday,$mon,$year);
  return $dateString;
}



sub addPackage{
  my ($this,$packobj) = @_;
  $this->removePackage($packobj);
  my $name = $packobj->getName();
  my $path = $packobj->getMainPath();
  my $newPackage = $this->newPackage($name,$path);
  $newPackage->{'Version'} = $packobj->getVersion();
  $newPackage->{'CheckSum'} = $packobj->getCheckSum();
  $newPackage->{'BuildString'} = $packobj->getBuildString();
  $newPackage->{'Date'} = generateDate
  $newPackage->{'Desc'} = $packobj->getDesc();
  $newPackage->{'MakeId'} = $packobj->getMakeId();
  $newPackage->{'Mode'} = $packobj->getMode();
  $newPackage->{'PackageVersion'} = $packobj->getPackageVersion();
  $newPackage->{'Size'} = $packobj->getSize();
  $newPackage->{'BuildString'} = $packobj->getBuildString();
  $newPackage->{'Valid'} = 1;
  $newPackage->{'TestFile'} = $packobj->getTestFile();
  $newPackage->{'FileList'} = $packobj->getFileList();
  my @paths = $packobj->GetPaths();
  $newPackage->{'Paths'} = \@paths;
  $newPackage->{'Profiles'} = $packobj->getProfiles();
  $newPackage->{'Require'} = $packobj->getRequire();
  $newPackage->{'Script'} = $packobj->getScript();
  $newPackage->{'SubPackages'} = [];
  return $newPackage;
}


sub createPackageObject{
  my ($this,$packname,$path,$instConfig) = @_;
  my %data  ;
  $path = normalize($path);
  $data{'PACKAGE_NAME'}=$packname;
  $data{'SOFTWARE_VERSION'}=$this->GetPackageValue($packname,$path,'Version');
  $data{'PACKAGE_VERSION'} = $this->GetPackageValue($packname,$path,'PackageVersion');
  $data{'MODE'} = $this->GetPackageValue($packname,$path,'Mode');
  $data{'REQUIRE'} = $this->GetPackageValue($packname,$path,'Require');
  $data{'DESC'} = $this->GetPackageValue($packname,$path,'Desc');
  $data{'SIZE'} =$this->GetPackageValue($packname,$path,'Size');
  $data{'BUILD_STRING'} = $this->GetPackageValue($packname,$path,'BuildString');
  $data{'MAKE_ID'} = $this->GetPackageValue($packname,$path,'MakeId');
  $data{'TEST_FILE'} = $this->GetPackageValue($packname,$path,'TestFile');
  my $packobj = SAPDB::Install::Templates2::InstUnit::new(\%data,$instConfig);
  $packobj->setMainPath($path);
  $packobj->setFileList($this->GetPackageValue($packname,$path,'FileList'));
  $packobj->setPaths($this->GetPackageValue($packname,$path,'Paths'));
  $packobj->setScript($this->GetPackageValue($packname,$path,'Script'));
  return $packobj;
}







sub removePackage{
  my ($this,$packobj) = @_;
  my $name = $packobj->getName();
  my $path = $packobj->getMainPath();
  $this->removePackageEntry($name,$path);
}

sub SetPackageValidFlag{
  my ($this,$pack,$flag)=@_;
  my $packname = $pack->getName();
  my $path = $pack->getMainPath();
  return $this->SetValidFlag($packname,$path,$flag);
}

sub SetValidFlag{
  my ($this,$packname,$path,$flag)=@_;
  return $this->SetPackageValue($packname,$path,'Valid',$flag);
#  my $value = ${%{${%{${%$hrRegistryData}{$packname}}}{$path}}}{'Valid'}=$flag;
}


sub removePackageEntry{
  my ($this,$packagename,$path)=@_;
  $path = normalize($path);
  my $dataroot=$this->getData();
  my $packageroot=$dataroot->{'HashRef_Packages'};
  my $hashref_paths=$packageroot->{$packagename};
  my $oldEntry = $hashref_paths->{$path};
#  print "Deleted: $path\n";
  delete $hashref_paths->{$path};
  return $oldEntry;
}

1;

